//
//      uncia - a big cat
//      Copyright (C) 2008 Peter Miller
//
//      This program is free software; you can redistribute it and/or
//      modify it under the terms of the GNU General Public License,
//      version 3, as published by the Free Software Foundation.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program. If not, see
//      <http://www.gnu.org/licenses/>.
//

#ifndef LIBUNCIA_INPUT_H
#define LIBUNCIA_INPUT_H

#include <boost/shared_ptr.hpp>

#include <libuncia/gcc_attributes.h>

namespace libuncia
{
    /**
      * The input class is used to represent an abstract input source.
      * It could be just about anything.
      *
      * The methods are tailored to being called from a streambuf
      * adapter.
      */
    class input
    {
    public:
        /**
          * The pointer typedef is provided to allow clients of this API
          * to be unaware of the actual details of the "smart" pointer
          * being used.  If it chages in future, all that is required is
          * a recompile.
          */
        typedef boost::shared_ptr<input> pointer;

        /**
          * The destructor.
          */
        virtual ~input();

    protected:
        /**
          * The default constructor.
          */
        input();

    public:
        /**
          * The read method may be used to read data from the input source.
          *
          * @param data
          *     Where the data is to be read into.
          * @param size
          *     The maximum number of bytes to be read.
          * @returns
          *     On success, the number of bytes read is returned (zero
          *     indicates end of file), and the file position is
          *     advanced by this number.  It is not an error if this
          *     number is smaller than the number of bytes requested;
          *     this may happen for example because fewer bytes are
          *     actually available right now.  On error, -1 is returned,
          *     and errno is set appropriately; in this case it is left
          *     unspecified whether the file position changes.
          */
        virtual ssize_t read(void *data, size_t size) = 0;

        /**
          * The fatal_error method may be used to report errors about
          * this input stream.  Where possible, the name of the input
          * will be inserted into the error message.
          *
          * @param fmt
          *     The text of the error message, with embedeed formatting
          *     controls which describe the remaining arguments.  See
          *     printf(3) for more information.
          */
        void fatal_error(const char *fmt, ...) const
                                                        FORMAT_PRINTF(2, 3);

        /**
          * The fatal_error method may be used to report errors about
          * this input stream.  Where possible, the name of the input
          * will be inserted into the error message.
          *
          * @param fmt
          *     The text of the error message, with embedeed formatting
          *     controls which describe the remaining arguments.  See
          *     vprintf(3) for more information.
          * @param ap
          *     Where to obtain the remaining arguments.
          */
        void fatal_error_v(const char *fmt, va_list ap) const
                                                        FORMAT_PRINTF(2, 0);

        /**
          * The filename_get method may be used to obtain the name of the
          * input.  This is used by the fatal_error methods to improve
          * error messages.
          */
        virtual std::string filename_get() const = 0;

    private:
        /**
          * The copy constructor.  Do not use.
          */
        input(const input &);

        /**
          * The assignment operator.  Do not use.
          */
        input &operator=(const input &);
    };
}

#endif // LIBUNCIA_INPUT_H
