//
// uncia - a big cat
// Copyright (C) 2008 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

#include <cerrno>
#include <map>

#include <libuncia/error_handler.h>
#include <libuncia/input/curl.h>


#define FATAL(function, reason) \
    libuncia::error_handler_get()->fatal_error("%s: %d: " function ": %s", \
        __FILE__, __LINE__, reason);


//
// If there is more than one URL open at a time, all are processed
// in parallel.  The multi-handle aggregates them all.
//
static CURLM *multi_handle;
static bool call_multi_immediate;
typedef std::map<CURL *, libuncia::input_curl *> stp_t;
typedef std::pair<CURL *, libuncia::input_curl *> row_t;
static stp_t stp;


libuncia::input_curl::~input_curl()
{
    //
    // Release libcurl resources.
    //
    curl_multi_remove_handle(multi_handle, handle);
    curl_easy_cleanup(handle);
    handle = 0;
    eof = true;

    //
    // Release dynamic memory resources.
    //
    delete [] curl_buffer;
    curl_buffer = 0;
    curl_buffer_position = 0;
    curl_buffer_length = 0;
    curl_buffer_maximum = 0;
}


//
// Libcurl calls this function when it receives more data.
//
static size_t
write_callback(char *data, size_t size, size_t nitems, void *p)
{
    libuncia::input_curl *icp = (libuncia::input_curl *)p;
    size_t nbytes = size * nitems;
    return icp->write_callback(data, nbytes);
}


libuncia::input_curl::input_curl(const std::string &arg) :
    fn(arg),
    pos(0),
    curl_buffer(0),
    curl_buffer_maximum(0),
    curl_buffer_position(0),
    curl_buffer_length(0),
    eof(false)
{
    handle = curl_easy_init();
    if (!handle)
        error_handler_get()->fatal_error("curl_easy_init");

    CURLcode err = curl_easy_setopt(handle, CURLOPT_ERRORBUFFER, errbuf);
    if (err)
        FATAL("curl_easy_setopt", curl_easy_strerror(err));

    err = curl_easy_setopt(handle, CURLOPT_URL, fn.c_str());
    if (err)
        FATAL("curl_easy_setopt", curl_easy_strerror(err));
    err = curl_easy_setopt(handle, CURLOPT_FILE, this);
    if (err)
        FATAL("curl_easy_setopt", curl_easy_strerror(err));
    err = curl_easy_setopt(handle, CURLOPT_VERBOSE, 0);
    if (err)
        FATAL("curl_easy_setopt", curl_easy_strerror(err));
    err = curl_easy_setopt(handle, CURLOPT_WRITEFUNCTION, ::write_callback);
    if (err)
        FATAL("curl_easy_setopt", curl_easy_strerror(err));
    err = curl_easy_setopt(handle, CURLOPT_FOLLOWLOCATION, 1);
    if (err)
        FATAL("curl_easy_setopt", curl_easy_strerror(err));

    if (!multi_handle)
    {
        multi_handle = curl_multi_init();
        if (!multi_handle)
            error_handler_get()->fatal_error("curl_multi_init");
    }
    CURLMcode merr = curl_multi_add_handle(multi_handle, handle);
    switch (merr)
    {
    case CURLM_CALL_MULTI_PERFORM:
        call_multi_immediate = true;
        break;

    case CURLM_OK:
        break;

    default:
        FATAL("curl_multi_add_handle", curl_multi_strerror(merr));
    }

    //
    // Start the fetch as soon as possible.
    //
    call_multi_immediate = true;

    //
    // Build an associate table from libcurl handles to our file pointers.
    //
    stp.insert(row_t(handle, this));
}


libuncia::input::pointer
libuncia::input_curl::create(const std::string &url)
{
    return pointer(new libuncia::input_curl(url));
}


size_t
libuncia::input_curl::write_callback(char *data, size_t nbytes)
{
    //
    // Grow the buffer if necessary.
    //
    // Always keep it a power of 2, because sigma(2**-n)==1, so we get
    // O(1) behaviour.  (That +32 means we are always just 32 bytes
    // short of a power of 2, leaving room for the malloc header, which
    // results in a nicer malloc fit on many systems.
    //
    if (curl_buffer_length + nbytes > curl_buffer_maximum)
    {
        for (;;)
        {
            curl_buffer_maximum = curl_buffer_maximum * 2 + 32;
            if (curl_buffer_length + nbytes <= curl_buffer_maximum)
                break;
        }
        char *new_curl_buffer = new char [curl_buffer_maximum];
        memcpy(new_curl_buffer, curl_buffer, curl_buffer_length);
        delete [] curl_buffer;
        curl_buffer = new_curl_buffer;
    }

    //
    // Copy the data into the buffer.
    //
    memcpy(curl_buffer + curl_buffer_length, data, nbytes);
    curl_buffer_length += nbytes;

    //
    // A negative return will stop the transfer for this stream.
    //
    return nbytes;
}


static libuncia::input_curl *
handle_to_fp(CURL *handle)
{
    assert(!stp.empty());
    stp_t::iterator it = stp.find(handle);
    if (it == stp.end())
    {
        libuncia::error_handler_get()->fatal_error
        (
            "%s: %d: handle %p has no mapping",
            __FILE__,
            __LINE__,
            handle
        );
    }
    return it->second;
}


/**
  * The perform function is a wrapper around the curl_multi_perform
  * function.  It checks for messages that may be waiting, waits in
  * select if necessary, and calls curl_multi_perform eventually.
  *
  * It is expected that this function will be repeatedly called from a
  * tight loop, so it doesn't loop itself.
  */

static void
perform(void)
{
    //
    // See if there are any messages waiting.
    // These tell us about errors, and completed transfers.
    //
    for (;;)
    {
        int msgs = 0;
        CURLMsg *msg = curl_multi_info_read(multi_handle, &msgs);
        if (!msg)
            break;
        if (msg->msg == CURLMSG_NONE)
            break;
        if (msg->msg != CURLMSG_DONE)
        {
            libuncia::error_handler_get()->fatal_error
            (
                "curl_multi_info_read -> %d (bug)",
                msg->msg
            );
        }
        libuncia::input_curl *fp = handle_to_fp(msg->easy_handle);
        fp->message(msg);
    }

    //
    // Look for more to happen.
    //
    if (call_multi_immediate)
    {
        call_multi_immediate = false;
        for (;;)
        {
            int num_xfer = 0;
            CURLMcode ret = curl_multi_perform(multi_handle, &num_xfer);
            switch (ret)
            {
            case CURLM_CALL_MULTI_PERFORM:
                call_multi_immediate = true;
                return;

            case CURLM_OK:
                return;

            default:
                libuncia::error_handler_get()->fatal_error
                (
                    "%s: %d: curl_multi_perform: %s",
                    __FILE__,
                    __LINE__,
                    curl_multi_strerror(ret)
                );
            }
        }
    }
    else
    {
        fd_set fdread;
        FD_ZERO(&fdread);
        fd_set fdwrite;
        FD_ZERO(&fdwrite);
        fd_set fdexcep;
        FD_ZERO(&fdexcep);

        // get file descriptors from the transfers
        int maxfd = 0;
        CURLcode err =
            (CURLcode)
            curl_multi_fdset(multi_handle, &fdread, &fdwrite, &fdexcep, &maxfd);
        if (err)
            FATAL("curl_multi_fdset", curl_easy_strerror(err));

        if (maxfd >= 0)
        {
            // set a suitable timeout to fail on
            struct timeval timeout;
            timeout.tv_sec = 60; // 1 minute
            timeout.tv_usec = 0;

            int rc = select(maxfd + 1, &fdread, &fdwrite, &fdexcep, &timeout);
            if (rc < 0)
            {
                if (errno != EINTR)
                {
                    libuncia::error_handler_get()->fatal_error_errno
                    (
                        "%s: %d: select",
                        __FILE__,
                        __LINE__
                    );
                    // NOTREACHED
                }
            }
            if (rc > 0)
            {
                //
                // Some sockets are ready.
                //
                call_multi_immediate = true;
            }
        }
    }
}


void
libuncia::input_curl::message(CURLMsg *msg)
{
    if (msg->data.result == 0)
    {
        // transfer over, no error
        eof = true;
    }
    else
    {
        libuncia::error_handler_get()->fatal_error
        (
            "read %s: %s",
            fn.c_str(),
            errbuf
        );
    }
}


ssize_t
libuncia::input_curl::read(void *data, size_t nbytes)
{
    //
    // attempt to fill buffer
    //
    while (!eof && curl_buffer_position + nbytes > curl_buffer_length)
        perform();

    //
    // Extract as much data as possible from the buffer.
    //
    size_t size_of_buffer = curl_buffer_length - curl_buffer_position;
    if (nbytes > size_of_buffer)
        nbytes = size_of_buffer;
    memcpy(data, curl_buffer + curl_buffer_position, nbytes);
    curl_buffer_position += nbytes;

    //
    // Rearrange the buffer so that it does not grow forever.
    //
    size_of_buffer = curl_buffer_length - curl_buffer_position;
    if (size_of_buffer == 0)
    {
        curl_buffer_position = 0;
        curl_buffer_length = 0;
    }
    else if (size_of_buffer <= curl_buffer_position)
    {
        // can shuffle the data down easily
        memcpy(curl_buffer, curl_buffer + curl_buffer_position, size_of_buffer);
        curl_buffer_position = 0;
        curl_buffer_length = size_of_buffer;
    }

    //
    // Return the number of bytes read.
    //
    pos += nbytes;
    return nbytes;
}


std::string
libuncia::input_curl::filename_get()
    const
{
    return fn;
}


bool
libuncia::input_curl::candidate(const std::string &file_name)
{
    const char *cp = file_name.c_str();
    if (!isalpha((unsigned char)*cp))
        return 0;
    for (;;)
    {
        ++cp;
        if (!isalpha((unsigned char)*cp))
            break;
    }
    return (cp[0] == ':' && cp[1] != '\0');
}
