//
// uncia - a big cat
// Copyright (C) 2003, 2005, 2006, 2008 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIBUNCIA_INPUT_CURL_H
#define LIBUNCIA_INPUT_CURL_H

#include <curl/curl.h>
#include <string>

#include <libuncia/input.h>

namespace libuncia
{
    /**
      * The input_curl class is used to represent an input via licurl.
      * Typically, a HTTP or FTP network file read.
      */
    class input_curl:
        public input
    {
    public:
        /**
          * The destructor.
          */
        virtual ~input_curl();

    private:
        /**
          * The constructor.
          * It is private on purpose, use the #create class method instead.
          *
          * @param url
          *     where to obtain the input to this filter.
          */
        input_curl(const std::string &url);

    public:
        /**
          * The create class method is used to create new dynmically
          * allocated instances of this class.
          *
          * @param url
          *     where to obtain the input to this filter.
          */
        static pointer create(const std::string &url);

        /**
          * The candidate class method is used to examine a filename and
          * see if it starts with a prototype name (e.g. http: or ftp:)
          * making it a likely candidate for the libuncia::input_curl class.
          */
        static bool candidate(const std::string &fn);

        // This is only public so that the CURL write callback can access it.
        size_t write_callback(char *data, size_t nbytes);

        // This is only public so that the CURL perform() can access it.
        void message(CURLMsg *msg);

    protected:
        // See base class for documentation.
        ssize_t read(void *data, size_t size);

        // See base class for documentation.
        std::string filename_get() const;

    private:
        CURL *handle;
        std::string fn;
        long pos;

        // buffer to store cached data
        char *curl_buffer;

        // currently allocated buffers length
        size_t curl_buffer_maximum;

        // start of data in buffer
        size_t curl_buffer_position;

        // end of data in buffer
        size_t curl_buffer_length;

        // end of file has been reached
        bool eof;

        char errbuf[CURL_ERROR_SIZE];

        /**
          * The default constructor.  Do not use.
          */
        input_curl();

        /**
          * The copy constructor.  Do not use.
          *
          * @param rhs
          *     The right hand side of the initialization.
          */
        input_curl(const input_curl &rhs);

        /**
          * The assignment operator.  Do not use.
          *
          * @param rhs
          *     The right hand side of the assignment.
          */
        input_curl &operator=(const input_curl &rhs);
    };
}

#endif // LIBUNCIA_INPUT_CURL_H
