//
// uncia - a big cat
// Copyright (C) 2008 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIBUNCIA_INPUT_FILTER_H
#define LIBUNCIA_INPUT_FILTER_H

#include <list>

#include <libuncia/input.h>
#include <libuncia/istream.h>

namespace libuncia
{
    /**
      * The input_filter abstreact class is used to represent a generic
      * input filter which provides everything except the actual read
      * method.
      */
    class input_filter:
        public input
    {
    public:
        /**
          * The destructor.
          */
        virtual ~input_filter();

    protected:
        /**
          * The constructor.
          * It is private on purpose, use the #create class method instead.
          *
          * @param deeper
          *     where to obtain the input to this filter.
          */
        input_filter(const pointer &deeper);

        /**
          * The deeper_get method is used to obtain the next character
          * from the deeper input source.
          *
          * @param c
          *     where to put the resuling character
          * @returns
          *     true on success, false on end-of-file or error
          */
        bool
        deeper_get(char &c)
        {
            if (holding_pen.empty())
                return deeper.get(c);
            c = holding_pen.back();
            holding_pen.pop_back();
            return true;
        }

        /**
          * The deeper_putback method is used to return data to the
          * deeper input stream.
          *
          * @param c
          *     the data to give back.  (Must not be std::eof.)
          * @note
          *     this can be thought of as using deeper.putback, but
          *     it actually uses its own holding area, in order to be
          *     able to have deeper push back than the std::istream
          *     guarantees.
          */
        void
        deeper_putback(char c)
        {
            holding_pen.push_back(c);
        }

        /**
          * The deeper_read method may be used to read data from the
          * deeper input source.
          *
          * @param data
          *     Where the data is to be read into.
          * @param size
          *     The maximum number of bytes to be read.
          * @returns
          *     On success, the number of bytes read is returned (zero
          *     indicates end of file), and the file position is
          *     advanced by this number.
          */
        ssize_t deeper_read(void *data, size_t size);

        /**
          * The deeper_getline method is used to obtain the next line
          * from the deeper input source (without the newline).
          *
          * @param result
          *     where to put the resuling text
          * @returns
          *     true on success, false on end-of-file or error
          */
        bool
        deeper_getline(std::string &result)
        {
            return deeper.getline(result);
        }

    protected:
        // See base class for documentation.
        std::string filename_get() const;

    private:
        /**
          * The deeper instance variable is used to remember where to
          * obtain the input to this filter.
          */
        libuncia::istream deeper;

        /**
          * The holding_pen instance variable is used to remember
          * characters which have been "returned" to the deeper input.
          */
        std::list<char> holding_pen;

        /**
          * The default constructor.  Do not use.
          */
        input_filter();

        /**
          * The copy constructor.  Do not use.
          *
          * @param rhs
          *     The right hand side of the initialization.
          */
        input_filter(const input_filter &rhs);

        /**
          * The assignment operator.  Do not use.
          *
          * @param rhs
          *     The right hand side of the assignment.
          */
        input_filter &operator=(const input_filter &rhs);
    };
}

#endif // LIBUNCIA_INPUT_FILTER_H
