//
// uncia - a big cat
// Copyright (C) 2008 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIBUNCIA_INPUT_FILTER_COLUMN_REMOVE_H
#define LIBUNCIA_INPUT_FILTER_COLUMN_REMOVE_H

#include <libuncia/input/filter.h>
#include <libuncia/istream.h>

namespace libuncia
{
    /**
      * The input_filter_column_remove class is used to represent an
      * input filter which may be used to remove character columns from
      * its input.
      */
    class input_filter_column_remove:
        public input_filter
    {
    public:
        /**
          * The destructor.
          */
        virtual ~input_filter_column_remove();

    private:
        /**
          * The constructor.
          * It is private on purpose, use the #create class method instead.
          *
          * @param deeper
          *     where to obtain the input to this filter.
          * @param left
          *     the left hand column, where to start removing, counting
          *     from zero
          * @param right
          *     the right hand column, where to stop removing, counting
          *     from zero (-1 == infinity)
          */
        input_filter_column_remove(const pointer &deeper, int left,
            int right = -1);

    public:
        /**
          * The create class method is used to create new dynmically
          * allocated instances of this class.
          *
          * @param deeper
          *     where to obtain the input to this filter.
          * @param left
          *     the left hand column, where to start removing, counting
          *     from zero
          * @param right
          *     the right hand column, where to stop removing, counting
          *     from zero (-1 == infinity)
          */
        static pointer create(const pointer &deeper, int left, int right = -1);

    protected:
        // See base class for documentation.
        ssize_t read(void *data, size_t size);

    private:
        /**
          * The left instance variable is used to remember the left hand
          * column, where to start removing, counting from zero.
          */
        unsigned left;

        /**
          * The right instance variable is used to remember the right
          * hand column, where to stop removing, counting from zero.
          */
        unsigned right;

        /**
          * The tab_width instance variable is used to remember the
          * column width of a horizontal TAB.
          */
        unsigned tab_width;

        /**
          * The column instance variable is used to remember the current
          * input column position, counting from zero.
          */
        unsigned column;

        /**
          * The default constructor.  Do not use.
          */
        input_filter_column_remove();

        /**
          * The copy constructor.  Do not use.
          *
          * @param rhs
          *     The right hand side of the initialization.
          */
        input_filter_column_remove(const input_filter_column_remove &rhs);

        /**
          * The assignment operator.  Do not use.
          *
          * @param rhs
          *     The right hand side of the assignment.
          */
        input_filter_column_remove &operator=(
            const input_filter_column_remove &rhs);
    };
}

#endif // LIBUNCIA_INPUT_FILTER_COLUMN_REMOVE_H
