//
// uncia - a big cat
// Copyright (C) 2008 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

#include <libuncia/input/filter/gunzip.h>
#include <libuncia/zlib_glue.h>

#ifndef Z_BUFSIZE
#  ifdef MAXSEG_64K
#    define Z_BUFSIZE 4096 // minimize memory usage for 16-bit DOS
#  else
#    define Z_BUFSIZE 16384
#  endif
#endif

// gzip flag byte
#define ASCII_FLAG   0x01 // bit 0 set: file probably ascii text
#define HEAD_CRC     0x02 // bit 1 set: header CRC present
#define EXTRA_FIELD  0x04 // bit 2 set: extra field present
#define ORIG_NAME    0x08 // bit 3 set: original file name present
#define COMMENT      0x10 // bit 4 set: file comment present
#define RESERVED     0xE0 // bits 5..7: reserved


libuncia::input_filter_gunzip::~input_filter_gunzip()
{
    int err = inflateEnd(&stream);
    if (err < 0)
        zlib_fatal_error(err);
    delete [] buf;
    buf = 0;
}


libuncia::input_filter_gunzip::input_filter_gunzip(const pointer &arg) :
    input_filter(arg),
    z_eof(false),
    crc(crc32(0L, Z_NULL, 0)),
    pos(0),
    buf(new Byte [Z_BUFSIZE])
{
    stream.zalloc = (alloc_func)0;
    stream.zfree = (free_func)0;
    stream.opaque = (voidpf)0;
    stream.next_in = Z_NULL;
    stream.avail_in = 0;
    stream.next_out = Z_NULL;
    stream.avail_out = 0;

    //
    // windowBits is passed < 0 to tell that there is no zlib header.
    // Note that in this case inflate *requires* an extra "dummy" byte
    // after the compressed stream in order to complete decompression
    // and return Z_STREAM_END. Here the gzip CRC32 ensures that 4
    // bytes are present after the compressed stream.
    //
    int err = inflateInit2(&stream, -MAX_WBITS);
    if (err < 0)
        zlib_fatal_error(err);

    //
    // Now read the file header.
    //
    read_header();
}


libuncia::input::pointer
libuncia::input_filter_gunzip::create(const pointer &a_deeper)
{
    return pointer(new input_filter_gunzip(a_deeper));
}


void
libuncia::input_filter_gunzip::zlib_fatal_error(int err)
{
    if (err >= 0)
        return;
    if (stream.msg)
        fatal_error("gunzip: %s (%s)", z_error(err), stream.msg);
    else
        fatal_error("gunzip: %s", z_error(err));
}


long
libuncia::input_filter_gunzip::getLong()
{
    long result = 0;
    for (int j = 0; j < 4; ++j)
    {
        unsigned char c = read_one_byte();
        result += c << (j * 8);
    }
    return result;
}


ssize_t
libuncia::input_filter_gunzip::read(void *data, size_t len)
{
    if (z_eof)
        return 0;

    Bytef *start = (Bytef *)data; // starting point for crc computation
    stream.next_out = (Bytef *)data;
    stream.avail_out = len;

    while (stream.avail_out > 0)
    {
        if (stream.avail_in == 0)
        {
            stream.next_in = buf;
            stream.avail_in = deeper_read(buf, Z_BUFSIZE);
            //
            // There should always be something left on the
            // input, because we have the CRC and Length
            // to follow.  Fatal error if not.
            //
            if (stream.avail_in <= 0)
            {
                fatal_error("gunzip: premature end of file");
            }
        }
        int err = inflate(&stream, Z_PARTIAL_FLUSH);
        if (err < 0)
            zlib_fatal_error(err);
        if (err == Z_STREAM_END)
        {
            z_eof = true;

            //
            // Push back the unused portion of the input stream.
            // (The way we wrote it, there shouldn't be much.)
            //
            while (stream.avail_in > 0)
            {
                stream.avail_in--;
                deeper_putback(stream.next_in[stream.avail_in]);
            }

            //
            // Fall out of the loop.
            //
            break;
        }
    }

    //
    // Calculate the running CRC
    //
    long result = stream.next_out - start;
    crc = crc32(crc, start, (uInt)result);

    //
    // Update the file position.
    //
    pos += result;

    //
    // At end-of-file we need to do some checking.
    //
    if (z_eof)
    {
        //
        // Check CRC
        //
        // Watch out for 64-bit machines.  This is what
        // those aparrently redundant 0xFFFFFFFF are for.
        //
        if (((unsigned long)getLong() & 0xFFFFFFFF) != (crc & 0xFFFFFFFF))
            fatal_error("gunzip: checksum mismatch");

        //
        // The uncompressed length here may be different
        // from pos in case of concatenated .gz
        // files.  But we don't write them that way,
        // so give an error if it happens.
        //
        // We shouldn't have 64-bit problems in this case.
        //
        if (getLong() != pos)
            fatal_error("gunzip: length mismatch");
    }

    //
    // Return success (failure always goes via input_format_error,
    // or zlib_fatal_error).
    //
    return result;
}


//
// Check the gzip header of a gz_stream opened for reading. Set the
// stream mode to transparent if the gzip magic header is not present;
// set err to Z_DATA_ERROR if the magic header is present but the
// rest of the header is incorrect.
//
// IN assertion: the stream this has already been created sucessfully;
// stream.avail_in is zero for the first time, but may be non-zero
// for concatenated .gz files.
//
static unsigned char gz_magic[2] = { 0x1F, 0x8B }; // gzip magic header


void
libuncia::input_filter_gunzip::read_header()
{
    //
    // Check for the magic number.
    // If it isn't present, assume transparent mode.
    //
    unsigned char c1 = read_one_byte();
    unsigned char c2 = read_one_byte();
    if (c1 != gz_magic[0] || c2 != gz_magic[1])
    {
        fatal_error
        (
            "gunzip: wrong magic number (got %02X%02X, expected %02X%02X)",
            c1,
            c2,
            gz_magic[0],
            gz_magic[1]
        );
    }

    //
    // Magic number present, now we require the rest of the header
    // to be present and correctly formed.
    //
    unsigned char method = read_one_byte();
    if (method != Z_DEFLATED)
        fatal_error("gunzip: not deflated encoding");
    unsigned char flags = read_one_byte();
    if ((flags & RESERVED) != 0)
        fatal_error("gunzip: unknown flags");

    // Discard time, xflags and OS code:
    for (uInt len = 0; len < 6; len++)
    {
        read_one_byte();
    }

    if (flags & EXTRA_FIELD)
    {
        // skip the extra field
        unsigned char e1 = read_one_byte();
        unsigned char e2 = read_one_byte();
        int elen = (e2 << 8) + e1;
        while (elen-- > 0)
            read_one_byte();
    }
    if (flags & ORIG_NAME)
    {
        // skip the original file name
        for (;;)
        {
            unsigned char c = read_one_byte();
            if (c == 0)
                break;
        }
    }
    if (flags & COMMENT)
    {
        // skip the .gz file comment
        for (;;)
        {
            unsigned char c = read_one_byte();
            if (c == 0)
                break;
        }
    }
    if (flags & HEAD_CRC)
    {
        // skip the header crc
        read_one_byte();
        read_one_byte();
    }
}


unsigned char
libuncia::input_filter_gunzip::read_one_byte()
{
    char c = 0;
    if (!deeper_get(c))
        fatal_error("gunzip: short file");
    return ((unsigned char)c);
}
