//
// uncia - a big cat
// Copyright (C) 2008 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIBUNCIA_INPUT_FILTER_UNIQUE_H
#define LIBUNCIA_INPUT_FILTER_UNIQUE_H

#include <libuncia/input/filter.h>
#include <libuncia/istream.h>

namespace libuncia
{
    /**
      * The input_filter_unique class is used to represent an input
      * filter which only passes throuth unique lines, and suppresses
      * duplicates.
      *
      * @note
      *     This filter does not detect repeated lines unless they are
      *     adjacent.  You may want to sort the input first.
      */
    class input_filter_unique:
        public input_filter
    {
    public:
        /**
          * The destructor.
          */
        virtual ~input_filter_unique();

    private:
        /**
          * The constructor.
          * It is private on purpose, use the #create class method instead.
          *
          * @param deeper
          *     where to obtain the input to this filter.
          * @param include_count
          *     whether or not to include the line count next to each
          *     output line.
          */
        input_filter_unique(const pointer &deeper, bool include_count = false);

    public:
        /**
          * The create class method is used to create new dynmically
          * allocated instances of this class.
          *
          * @param deeper
          *     where to obtain the input to this filter.
          * @param include_count
          *     whether or not to include the line count next to each
          *     output line.
          */
        static pointer create(const pointer &deeper,
            bool include_count = false);

    protected:
        // See base class for documentation.
        ssize_t read(void *data, size_t size);

    private:
        /**
          * The include_count instance variable is used to remember
          * whether or not to include the line count next to each output
          * line.
          */
        bool include_count;

        /**
          * The deeper instance variable is used to remember the text of
          * the previous line.
          */
        std::string previous_line;

        /**
          * The repeat_count instance variable is used to remember the
          * number of times the previous line has occurred.
          */
        long repeat_count;

        /**
          * The emit instance variable is used to remember the next
          * remaining to pass upstream.
          */
        std::string emit;

        /**
          * The emit_previous_line instance variable is used to move the
          * text (if any) to be passed upstream, via the #emit buffer.
          */
        void emit_previous_line();

        /**
          * The default constructor.  Do not use.
          */
        input_filter_unique();

        /**
          * The copy constructor.  Do not use.
          *
          * @param rhs
          *     The right hand side of the initialization.
          */
        input_filter_unique(const input_filter_unique &rhs);

        /**
          * The assignment operator.  Do not use.
          *
          * @param rhs
          *     The right hand side of the assignment.
          */
        input_filter_unique &operator=(const input_filter_unique &rhs);
    };
}

#endif // LIBUNCIA_INPUT_FILTER_UNIQUE_H
