//
// uncia - a big cat
// Copyright (C) 2008 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIBUNCIA_INPUT_PASTE_H
#define LIBUNCIA_INPUT_PASTE_H

#include <libuncia/input.h>
#include <libuncia/istream.h>

namespace libuncia
{
    /**
      * The input_paste class is used to represent an input filter which
      * glues two inputs together line-by-line.
      */
    class input_paste:
        public input
    {
    public:
        /**
          * The destructor.
          */
        virtual ~input_paste();

    private:
        /**
          * The constructor.
          * It is private on purpose, use the #create class method instead.
          *
          * @param deeper1
          *     where to obtain the left hand input to this filter.
          * @param delimiter
          *     strings to separate the two columns
          * @param deeper2
          *     where to obtain the right hand input to this filter.
          */
        input_paste(const pointer &deeper1, const std::string &delimiter,
            const pointer &deeper2);

    public:
        /**
          * The create class method is used to create new dynmically
          * allocated instances of this class.  It puts a TAB between
          * the columns.
          *
          * @param deeper1
          *     where to obtain the left hand input to this filter.
          * @param deeper2
          *     where to obtain the right hand input to this filter.
          */
        static pointer create(const pointer &deeper1, const pointer &deeper2);

        /**
          * The create class method is used to create new dynmically
          * allocated instances of this class.
          *
          * @param delimiter
          *     strings to separate the two columns
          * @param deeper1
          *     where to obtain the left hand input to this filter.
          * @param deeper2
          *     where to obtain the right hand input to this filter.
          */
        static pointer create(const std::string &delimiter,
            const pointer &deeper1, const pointer &deeper2);

    protected:
        // See base class for documentation.
        ssize_t read(void *data, size_t size);

        // See base class for documentation.
        std::string filename_get() const;

    private:
        /**
          * The deeper1 instance variable is used to remember where to
          * obtain the left-hand input to this filter.
          */
        libuncia::istream deeper1;

        /**
          * The delimiter instance variable is used to remember what to
          * put between the columns of output.
          */
        std::string delimiter;

        /**
          * The deeper2 instance variable is used to remember where to
          * obtain the right-hand input to this filter.
          */
        libuncia::istream deeper2;

        /**
          * The buffer instance variable is used to buffer the output
          * line when that lines spans #read calls.
          */
        std::string buffer;

        /**
          * The default constructor.  Do not use.
          */
        input_paste();

        /**
          * The copy constructor.  Do not use.
          *
          * @param rhs
          *     The right hand side of the initialization.
          */
        input_paste(const input_paste &rhs);

        /**
          * The assignment operator.  Do not use.
          *
          * @param rhs
          *     The right hand side of the assignment.
          */
        input_paste &operator=(const input_paste &rhs);
    };
}

#endif // LIBUNCIA_INPUT_PASTE_H
