//
//      uncia - a big cat
//      Copyright (C) 2008 Peter Miller
//
//      This program is free software; you can redistribute it and/or
//      modify it under the terms of the GNU General Public License,
//      version 3, as published by the Free Software Foundation.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program. If not, see
//      <http://www.gnu.org/licenses/>.
//

#ifndef LIBUNCIA_ISTREAM_H
#define LIBUNCIA_ISTREAM_H

#include <cstdarg>
#include <iostream>

#include <libuncia/gcc_attributes.h>
#include <libuncia/input.h>

namespace libuncia
{
    /**
      * The istream class is used to represent an adapter between C++
      * standard istream and libuncia::input
      */
    class istream:
        public std::istream
    {
    public:
        /**
          * The destructor.
          */
        virtual ~istream();

        /**
          * The constructor.
          *
          * @param rhs
          *     The input source to be consumed.
          */
        istream(const input::pointer &rhs);

        /**
          * The fatal_error method may be used to report errors about
          * this input stream.  Where possible, the name of the input
          * will be inserted into the error message.
          *
          * @param fmt
          *     The text of the error message, with embedeed formatting
          *     controls which describe the remaining arguments.  See
          *     printf(3) for more information.
          */
        void fatal_error(const char *fmt, ...)
                                                        FORMAT_PRINTF(2, 3);

        /**
          * The fatal_error method may be used to report errors about
          * this input stream.  Where possible, the name of the input
          * will be inserted into the error message.
          *
          * @param fmt
          *     The text of the error message, with embedeed formatting
          *     controls which describe the remaining arguments.  See
          *     vprintf(3) for more information.
          * @param ap
          *     Where to obtain the remaining arguments.
          */
        void fatal_error_v(const char *fmt, va_list ap)
                                                        FORMAT_PRINTF(2, 0);

        /**
          * The filename_get method may be used to obtain the name of
          * the input.  This is used by the fatal_error methods to
          * improve error messages.
          */
        std::string filename_get() const;

        /**
          * The getline method may be used to read a line of input.
          * The newline is read, but <b>not</b> placed in the result.
          *
          * @param result
          *     where to put the text of the line, sans newline.
          * @returns
          *     true if a line heas been read, false on end-of-file or
          *     deeper error.
          */
        bool getline(std::string &result);

    private:
        /**
          * The raw instance variable is used to remember where the
          * output is sent.  This permits it to be safely disposed of
          * by the destructor, and error messages to be passed down the
          * filter chain.
          */
        libuncia::input::pointer raw;

        /**
          * The default constructor.  Do not use.
          */
        istream();

        /**
          * The copy constructor.  Do not use.
          */
        istream(const istream &);

        /**
          * The assignment operator.  Do not use.
          */
        istream &operator=(const istream &);
    };
}

#endif // LIBUNCIA_ISTREAM_H
