//
//      uncia - a big cat
//      Copyright (C) 2008 Peter Miller
//
//      This program is free software; you can redistribute it and/or
//      modify it under the terms of the GNU General Public License,
//      version 3, as published by the Free Software Foundation.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program. If not, see
//      <http://www.gnu.org/licenses/>.
//

#ifndef LIBUNCIA_OUTPUT_H
#define LIBUNCIA_OUTPUT_H

#include <boost/shared_ptr.hpp>
#include <libuncia/gcc_attributes.h>

namespace libuncia
{
    /**
      * The output class is used to represent an abstract output sink.  It
      * could be just about anything.
      *
      * The methods are tailored to being called from a streambuf adapter.
      */
    class output
    {
    public:
        /**
          * The pointer typedef is provided to allow clients of this API
          * to be unaware of the actual details of the "smart" pointer
          * being used.  If it chages in future, all that is required is a
          * recompile.
          */
        typedef boost::shared_ptr<output> pointer;

        /**
          * The destructor.
          */
        virtual ~output();

    protected:
        /**
          * The default constructor.
          */
        output();

    public:
        /**
          * The write method is used to append data to the output.
          *
          * @param data
          *     The data to be written.
          * @param size
          *     The number of bytes of data to be written.
          * @returns
          *     On success, the number of bytes written is returned (zero
          *     indicates nothing was written).  On error, -1 is returned,
          *     and errno is set appropriately.
          */
        virtual ssize_t write(const void *data, size_t size) = 0;

        /**
          * The sync method is used to ensure the data has been written to
          * the output.
          */
        virtual int sync() = 0;

        /**
          * The fatal_error method may be used to report errors about
          * this output stream.  Where possible, the name of the output
          * will be inserted into the error message.
          *
          * @param fmt
          *     The text of the error message, with embedeed formatting
          *     controls which describe the remaining arguments.  See
          *     printf(3) for more information.
          */
        void fatal_error(const char *fmt, ...) const
                                                        FORMAT_PRINTF(2, 3);

        /**
          * The fatal_error method may be used to report errors about
          * this output stream.  Where possible, the name of the output
          * will be inserted into the error message.
          *
          * @param fmt
          *     The text of the error message, with embedeed formatting
          *     controls which describe the remaining arguments.  See
          *     vprintf(3) for more information.
          * @param ap
          *     Where to obtain the remaining arguments.
          */
        void fatal_error_v(const char *fmt, va_list ap) const
                                                        FORMAT_PRINTF(2, 0);

        /**
          * The filename_get method may be used to obtain the name of the
          * output.  This is used by the fatal_error methods to improve
          * error messages.
          */
        virtual std::string filename_get() const = 0;

    private:
        /**
          * The copy constructor.  Do not use.
          */
        output(const output &);

        /**
          * The assignment operator.  Do not use.
          */
        output &operator=(const output &);
    };
}

#endif // LIBUNCIA_OUTPUT_H
