//
// uncia - a big cat
// Copyright (C) 2008 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

#include <libuncia/output/filter/fold.h>


libuncia::output_filter_fold::~output_filter_fold()
{
    fold(0);
}


static inline size_t
fix_hanging(size_t line_width, int a_hanging)
{
    if (a_hanging < 0)
        return 0;
    size_t hanging(a_hanging);
    if (hanging >= line_width)
        return 0;
    return hanging;
}


libuncia::output_filter_fold::output_filter_fold(const pointer &a_deeper,
        bool a_spaces, int a_line_width, int a_tab_width, int a_hanging) :
    output_filter(a_deeper),
    spaces(a_spaces),
    line_width(a_line_width > 0 ? a_line_width : 75),
    tab_width(a_tab_width > 0 ? a_tab_width : 8),
    hanging(fix_hanging(line_width, a_hanging)),
    column(0),
    line_highwater(0),
    col_highwater(0)
{
}


libuncia::output::pointer
libuncia::output_filter_fold::create(const pointer &a_deeper, bool a_spaces,
    int a_line_width, int a_tab_width, int a_hanging)
{
    return
        pointer
        (
            new output_filter_fold
            (
                a_deeper,
                a_spaces,
                a_line_width,
                a_tab_width,
                a_hanging
            )
        );
}


ssize_t
libuncia::output_filter_fold::write(const void *data, size_t size)
{
    const char *start = (const char *)data;
    const char *end = start + size;
    const char *cp = start;
    while (cp < end)
    {
        unsigned char c = *cp++;
        switch (c)
        {
        case '\f':
        case '\n':
        case '\v':
            fold(c);
            break;

        case '\r':
            column = 0;
            break;

        case '\t':
            assert(tab_width >= 1);
            column = ((column / tab_width) + 1) * tab_width;
            break;

        case ' ':
            ++column;
            break;

        case '\b':
            if (column > 0)
                --column;
            break;

        default:
            // assume non printing characters are one column wide.
            while (line.size() <= column)
                line.push_back(col_t());
            col_t &col_r(line[column]);
            col_r.push_back(c);
            ++column;
            if (column > line_highwater)
                line_highwater = column;
            if (col_r.size() > col_highwater)
                col_highwater = col_r.size();
            break;
        }
    }
    return size;
}


void
libuncia::output_filter_fold::fold(char newline)
{
    if (!newline && !line_highwater)
        return;
    if (!line_highwater)
    {
        deeper_putc('\n');
        return;
    }
    size_t max_lw = line_width;
    size_t part = 0;
    while (part < line_highwater)
    {
        size_t opos = 0;
        size_t lw = max_lw;
        if (part + lw > line_highwater)
            lw = line_highwater - part;
        if (spaces && part + lw < line_highwater && !line[part + lw].empty())
        {
            size_t old_lw = lw;
            while (lw > 0 && line[part + lw - 1].size())
                --lw;
            if (lw == 0)
                lw = old_lw;
        }
        for (size_t y = 0; y < col_highwater; ++y)
        {
            for (size_t x = part; x < part + lw && x < line_highwater; ++x)
            {
                col_t &col_r(line[x]);
                if (y >= col_r.size())
                    continue;
                size_t x_col = x - part;
                if (opos > x_col)
                {
                    deeper_putc('\r');
                    opos = 0;
                }
                char c = col_r[y];

                // adjust for hanging indent
                if (part == 0 && x == 0)
                    max_lw -= hanging;
                else
                {
                    if (opos == 0)
                    {
                        for (size_t s = 0; s < hanging; ++s)
                        {
                            deeper_putc(' ');
                        }
                    }
                }

                // move to the correct column
                while (opos < x_col)
                {
                    deeper_putc(' ');
                    ++opos;
                }

                // emit the character
                deeper_putc(c);
                ++opos;
            }
        }
        deeper_putc('\n');
        opos = 0;
        part += lw;
        if (spaces)
        {
            while (part < line_highwater && line[part].empty())
            {
                ++part;
            }
        }
    }

    //
    // clear out the line buffer
    //
    for (line_t::iterator it = line.begin(); it != line.end(); ++it)
        it->clear();
    line_highwater = 0;
    col_highwater = 0;

    //
    // return to the start of the line
    //
    column = 0;
}


#include <sys/ioctl.h>
#include <termios.h>


#define MIN_TERM_WIDTH 20
#define MAX_TERM_WIDTH 32767
#define DEFAULT_TERM_WIDTH 80


int
libuncia::output_filter_fold::get_terminal_width()
{
    static int terminal_width;

    //
    // Use environment variable, if given.
    //
    if (!terminal_width)
    {
        const char *cp = getenv("COLS");
        if (cp && *cp)
        {
            int n = atoi(cp);
            if (n > 0)
            {
                if (n < MIN_TERM_WIDTH)
                    n = MIN_TERM_WIDTH;
                if (n > MAX_TERM_WIDTH)
                    n = MAX_TERM_WIDTH;
                terminal_width = n;
            }
        }
    }

    //
    // Use stdout attributes, if available.
    //
#ifdef TIOCGWINSZ
    if (!terminal_width)
    {
        struct winsize  window_size;

        if (ioctl(0, TIOCGWINSZ, &window_size) == 0)
        {
            if (!terminal_width && window_size.ws_col > 0)
            {
                terminal_width = window_size.ws_col;
                if (terminal_width < MIN_TERM_WIDTH)
                    terminal_width = MIN_TERM_WIDTH;
                if (terminal_width > MAX_TERM_WIDTH)
                    terminal_width = MAX_TERM_WIDTH;
            }
        }
    }
#endif

    //
    // Use defaults if all else fails.
    //
    if (!terminal_width)
        terminal_width = DEFAULT_TERM_WIDTH;

    return terminal_width;
}
