//
// uncia - a big cat
// Copyright (C) 2008 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIBUNCIA_OUTPUT_FILTER_FOLD_H
#define LIBUNCIA_OUTPUT_FILTER_FOLD_H

#include <vector>
#include <libuncia/output/filter.h>

namespace libuncia
{
    /**
      * The output_filter_fold class is used to represent an output
      * filter which folds long lines.  It even copes with overwriting
      * using carriage returns and/or back space to overwrite character
      * positions.
      */
    class output_filter_fold:
        public output_filter
    {
    public:
        /**
          * The destructor.
          */
        virtual ~output_filter_fold();

    private:
        /**
          * The constructor.
          * It is private on purpose, use the #create class method instead.
          *
          * @param deeper
          *     where to write the output of this filter
          * @param spaces
          *     whether to fold on spaces (true), or simply cut off
          *     lines where the break occurs (false)
          * @param line_width
          *     the desired width of the output
          * @param tab_width
          *     the width of an input tab
          * @param hanging
          *     depth of hanging indent
          */
        output_filter_fold(const pointer &deeper, bool spaces = true,
            int line_width = 75, int tab_width = 8, int hanging = 0);

    public:
        /**
          * The create class method is used to create new dynmically
          * allocated instances of this class.
          *
          * @param deeper
          *     where to write the output of this filter
          * @param spaces
          *     whether to fold on spaces (true), or simply cut off
          *     lines where the break occurs (false)
          * @param line_width
          *     the desired width of the output
          * @param tab_width
          *     the width of an input tab
          * @param hanging
          *     depth of hanging indent
          */
        static pointer create(const pointer &deeper, bool spaces = true,
            int line_width = 75, int tab_width = 8, int hanging = 0);

        /**
          * The get terminal width class method is used to determine the
          * width of the startdard output, if it is going to a terminal
          * or terminal emulator.  Defaults to 80 if it can't figure it
          * out.
          */
        static int get_terminal_width();

    protected:
        // See base class for documentation.
        ssize_t write(const void *data, size_t size);

    private:
        /**
          * The spaces
          * instance variable is used to remember
          * whether to fold on spaces (true), or simply cut
          * off lines where the break occurs (false)
          */
        bool spaces;

        /**
          * The line_width instance variable is used to remember the
          * desired width of the output.
          */
        size_t line_width;

        /**
          * The tab_width instance variable is used to remember the
          * width of an input tab.
          */
        size_t tab_width;

        /**
          * The hanging instance variable is used to remember depth of
          * hanging indent.
          */
        size_t hanging;

        typedef std::vector<char> col_t;

        typedef std::vector<col_t> line_t;

        line_t line;

        size_t column;

        size_t line_highwater;

        size_t col_highwater;

        /**
          * The fold method is used by the #write method to write the
          * lines to the output, splitting where required.
          */
        void fold(char newline);

        /**
          * The default constructor.  Do not use.
          */
        output_filter_fold();

        /**
          * The copy constructor.  Do not use.
          *
          * @param rhs
          *     The right hand side of the initialization.
          */
        output_filter_fold(const output_filter_fold &rhs);

        /**
          * The assignment operator.  Do not use.
          *
          * @param rhs
          *     The right hand side of the assignment.
          */
        output_filter_fold &operator=(const output_filter_fold &rhs);
    };
}

#endif // LIBUNCIA_OUTPUT_FILTER_FOLD_H
