//
// uncia - a big cat
// Copyright (C) 2008 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

#include <libuncia/output/filter/gzip.h>


#ifndef Z_BUFSIZE
#ifdef MAXSEG_64K
#define Z_BUFSIZE 4096 // minimize memory usage for 16-bit DOS
#else
#define Z_BUFSIZE 16384
#endif
#endif

static unsigned char gzip_magic[2] = { 0x1F, 0x8B }; // gzip magic header


void
libuncia::output_filter_gzip::drop_dead(int err)
{
    fatal_error("gzip: %s", z_error(err));
}


void
libuncia::output_filter_gzip::output_long_le(uLong x)
{
    for (int n = 0; n < 4; n++)
    {
        deeper_putc((unsigned char)x);
        x >>= 8;
    }
}


libuncia::output_filter_gzip::~output_filter_gzip()
{
    //
    // finish sending the compressed stream
    //
    stream.avail_in = 0; // should be zero already anyway
    if (stream.avail_out == 0)
    {
        deeper_write(outbuf, Z_BUFSIZE);
        stream.next_out = outbuf;
        stream.avail_out = Z_BUFSIZE;
    }
    for (;;)
    {
        int err = deflate(&stream, Z_FINISH);
        if (err < 0)
            drop_dead(err);
        uInt len = Z_BUFSIZE - stream.avail_out;
        if (!len)
            break;
        deeper_write(outbuf, len);
        stream.next_out = outbuf;
        stream.avail_out = Z_BUFSIZE;
    }

    //
    // and the trailer
    //
    output_long_le(crc);
    output_long_le(stream.total_in);

    //
    // Clean up any resources we were using.
    //
    if (stream.state != NULL)
        deflateEnd(&stream);
    delete [] outbuf;
    outbuf = 0;
}


libuncia::output_filter_gzip::output_filter_gzip(const output::pointer &arg,
        int level) :
    output_filter(arg),
    outbuf(new Byte [Z_BUFSIZE]),
    crc(0),
    pos(0),
    bol(true)
{
    crc = crc32(0L, Z_NULL, 0);
    stream.avail_in = 0;
    stream.avail_out = 0;
    stream.next_in = NULL;
    stream.next_out = NULL;
    stream.opaque = (voidpf)0;
    stream.zalloc = (alloc_func)0;
    stream.zfree = (free_func)0;

    if (level < Z_BEST_SPEED)
        level = Z_BEST_SPEED;
    else if (level > Z_BEST_COMPRESSION)
        level = Z_BEST_COMPRESSION;

    //
    // Set the parameters for the compression.
    // Note: windowBits is passed < 0 to suppress zlib header.
    //
    int err =
        deflateInit2
        (
            &stream,
            level,
            Z_DEFLATED,         // method
            -MAX_WBITS,         // windowBits
            DEF_MEM_LEVEL,      // memLevel
            Z_DEFAULT_STRATEGY  // strategy
        );
    if (err != Z_OK)
        drop_dead(err);

    stream.next_out = outbuf;
    stream.avail_out = Z_BUFSIZE;

    //
    // Write a very simple .gz header:
    //
    deeper_putc(gzip_magic[0]);
    deeper_putc(gzip_magic[1]);
    deeper_putc(Z_DEFLATED);
    deeper_putc(0); // flags
    output_long_le(0L); // time
    deeper_putc(0); // xflags
    deeper_putc(3); // always use unix OS_CODE
}


libuncia::output::pointer
libuncia::output_filter_gzip::create(const output::pointer &a_deeper, int level)
{
    return pointer(new output_filter_gzip(a_deeper, level));
}


ssize_t
libuncia::output_filter_gzip::write(const void *buf, size_t len)
{
    if (len > 0)
        bol = (((const char *)buf)[len - 1] == '\n');
    stream.next_in = (Bytef *)buf;
    stream.avail_in = len;
    while (stream.avail_in != 0)
    {
        if (stream.avail_out == 0)
        {
            deeper_write(outbuf, Z_BUFSIZE);
            stream.next_out = outbuf;
            stream.avail_out = Z_BUFSIZE;
        }
        int err = deflate(&stream, Z_NO_FLUSH);
        if (err != Z_OK)
            drop_dead(err);
    }
    crc = crc32(crc, (Bytef *)buf, len);
    pos += len;
    return len;
}
