//
// uncia - a big cat
// Copyright (C) 2008 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIBUNCIA_OUTPUT_FILTER_GZIP_H
#define LIBUNCIA_OUTPUT_FILTER_GZIP_H

#include <libuncia/output/filter.h>
#include <libuncia/zlib_glue.h>

namespace libuncia
{
    /**
      * The output_filter_gzip class is used to represent an output
      * filter which compresses its output using the gzip(1) algorithm.
      */
    class output_filter_gzip:
        public output_filter
    {
    public:
        /**
          * The destructor.
          */
        virtual ~output_filter_gzip();

    private:
        /**
          * The constructor.
          * It is private on purpose, use the #create class method instead.
          *
          * @param deeper
          *     where to send the output of this filter.
          * @param level
          *     the compression factor (1 best speed .. 9 best compression)
          */
        output_filter_gzip(const pointer &deeper, int level);

    public:
        /**
          * The create class method is used to create new dynmically
          * allocated instances of this class.
          *
          * @param deeper
          *     where to send the output of this filter.
          * @param level
          *     the compression factor (1 best speed .. 9 best compression)
          */
        static pointer create(const pointer &deeper, int level = 9);

    protected:
        // See base class for documentation.
        ssize_t write(const void *data, size_t size);

    private:
        /**
          * The stream instance variable is used to remember somthign
          * that the gzip code wants to remember.  It is opaque to us.
          */
        z_stream stream;

        /**
          * The outbuf instance variable is used to remember the
          * buffered output data.
          */
        Byte *outbuf;

        /**
          * The crc instance variable is used to remember the crc32 of
          * the uncompressed data.
          */
        uLong crc;

        /**
          * The pos instance variable is used to remember the current
          * output position.
          */
        long pos;

        /**
          * The bol instance variable is used to remember whether or not
          * the output is positioned at the beginning of a lone.
          */
        bool bol;

        /**
          * The drop_dead method is used to report a fatal error from
          * the gzip engine.
          */
        void drop_dead(int err);

        /**
          * The output_long_le method is used to write a little endian
          * 4-byte value to the deeper output stream.
          */
        void output_long_le(uLong x);

        /**
          * The default constructor.  Do not use.
          */
        output_filter_gzip();

        /**
          * The copy constructor.  Do not use.
          *
          * @param rhs
          *     The right hand side of the initialization.
          */
        output_filter_gzip(const output_filter_gzip &rhs);

        /**
          * The assignment operator.  Do not use.
          *
          * @param rhs
          *     The right hand side of the assignment.
          */
        output_filter_gzip &operator=(const output_filter_gzip &rhs);
    };
}

#endif // LIBUNCIA_OUTPUT_FILTER_GZIP_H
