//
// uncia - a big cat
// Copyright (C) 2001-2006, 2008 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

#include <libuncia/output/filter/quoted_printable.h>


libuncia::output_filter_quoted_printable::~output_filter_quoted_printable()
{
    while (pos)
        eoln_partial();
}


libuncia::output_filter_quoted_printable::output_filter_quoted_printable(
        const pointer &a_deeper) :
    output_filter(a_deeper),
    allow_international_characters(false),
    pos(0)
{
}


libuncia::output::pointer
libuncia::output_filter_quoted_printable::create(const pointer &a_deeper)
{
    return pointer(new output_filter_quoted_printable(a_deeper));
}


static inline char
map_hex(int x)
{
    static char hex_digit[16] =
    {
        '0', '1', '2', '3', '4', '5', '6', '7',
        '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'
    };
    return hex_digit[x & 15];
}


void
libuncia::output_filter_quoted_printable::eoln(bool soft)
{
    glyph_t *cp = glyph;
    glyph_t *end = cp + pos;
    for (; cp < end; ++cp)
    {
        unsigned char c = cp->text;
        if (cp->quote_it)
        {
            deeper_putc('=');
            deeper_putc(map_hex(c >> 4));
            deeper_putc(map_hex(c));
        }
        else
            deeper_putc(c);
    }
    if (soft)
        deeper_putc('=');
    deeper_putc('\n');
    pos = 0;
}


void
libuncia::output_filter_quoted_printable::eoln_partial()
{
    //
    // The line is too long.  We need to back up a few
    // characters.  We must allow one column for the '='
    // soft line break
    // (which is why we say >=MAX_LINE_LEN instead of >MAX_LINE_LEN).
    //
    long oldpos = pos;
    long newpos = pos;
    while (newpos > 0 && glyph[newpos - 1].cumulative >= MAX_LINE_LEN)
        --newpos;
    long newpos_max = newpos;

    //
    // It's worth hunting for a white space character, it looks nicer.
    //
    while
    (
        newpos > 0
    &&
        glyph[newpos - 1].text != ' '
    &&
        glyph[newpos - 1].text != '\t'
    )
        --newpos;
    if (newpos == 0)
        newpos = newpos_max;

    //
    // re-write the line length, and emit the partial line.
    //
    pos = newpos;
    eoln(true);

    //
    // Move everything down.
    //
    for (int j = 0; j + newpos < oldpos; ++j)
    {
        glyph_t *gp = glyph + j;
        *gp = glyph[newpos + j];
        gp->cumulative = (j ? gp[-1].cumulative : 0) + gp->width;
    }
    pos = oldpos - newpos;
}


void
libuncia::output_filter_quoted_printable::eoln_hard()
{
    //
    // We are required to quote trailing spaces or tabs.
    //
    if (pos)
    {
        glyph_t *gp = glyph + pos - 1;
        if (gp->text == ' ' || gp->text == '\t')
        {
            gp->quote_it = true;
            gp->cumulative += 3 - gp->width;
            gp->width = 3;

            //
            // This could make the line longer than the
            // maximum, in which case we need to emit the
            // partial line first.
            //
            if (gp->cumulative > MAX_LINE_LEN)
                eoln_partial();
        }
    }

    //
    // now emit the whole lot.
    //
    eoln(false);
}


ssize_t
libuncia::output_filter_quoted_printable::write(const void *data, size_t size)
{
    const unsigned char *start = (const unsigned char *)data;
    const unsigned char *end = start + size;
    const unsigned char *cp = start;
    while (cp < end)
    {
        unsigned char c = *cp++;
        if (c == '\n')
        {
            eoln_hard();
            continue;
        }

        glyph_t *gp = glyph + pos;
        gp->text = c;
        gp->width = 1;
        gp->quote_it = false;
        gp->cumulative = 0;
        switch (c)
        {
        case '=':
            gp->width = 3;
            gp->quote_it = true;
            break;

        case '\t':
            {
                int col1 = (pos ? gp[-1].cumulative : 0);
                int col2 = (col1 + 8) & ~7;
                gp->width = col2 - col1;
            }
            break;

        case ' ':
        case '!':
        case '"':
        case '#':
        case '$':
        case '%':
        case '&':
        case '\'':
        case '(':
        case ')':
        case '*':
        case '+':
        case ',':
        case '-':
        case '.':
        case '/':
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
        case ':':
        case ';':
        case '<':
        case '>':
        case '?':
        case '@':
        case 'A':
        case 'B':
        case 'C':
        case 'D':
        case 'E':
        case 'F':
        case 'G':
        case 'H':
        case 'I':
        case 'J':
        case 'K':
        case 'L':
        case 'M':
        case 'N':
        case 'O':
        case 'P':
        case 'Q':
        case 'R':
        case 'S':
        case 'T':
        case 'U':
        case 'V':
        case 'W':
        case 'X':
        case 'Y':
        case 'Z':
        case '[':
        case '\\':
        case ']':
        case '^':
        case '_':
        case '`':
        case 'a':
        case 'b':
        case 'c':
        case 'd':
        case 'e':
        case 'f':
        case 'g':
        case 'h':
        case 'i':
        case 'j':
        case 'k':
        case 'l':
        case 'm':
        case 'n':
        case 'o':
        case 'p':
        case 'q':
        case 'r':
        case 's':
        case 't':
        case 'u':
        case 'v':
        case 'w':
        case 'x':
        case 'y':
        case 'z':
        case '{':
        case '|':
        case '}':
        case '~':
            break;

        default:
            // C locale
            if
            (
                allow_international_characters
            ?
                ((c & 0x7F) < ' ')
            :
                !isprint(c)
            )
            {
                gp->width = 3;
                gp->quote_it = true;
            }
            break;
        }
        gp->cumulative = (pos ? gp[-1].cumulative : 0) + gp->width;
        ++pos;

        if (gp->cumulative > MAX_LINE_LEN)
            eoln_partial();
    }
    return size;
}
