//
// uncia - a big cat
// Copyright (C) 2001, 2002, 2005, 2006, 2008 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIBUNCIA_OUTPUT_FILTER_QUOTED_PRINTABLE_H
#define LIBUNCIA_OUTPUT_FILTER_QUOTED_PRINTABLE_H

#include <libuncia/output/filter.h>

namespace libuncia
{
    /**
      * The output_filter_quoted_printable class is used to represent an output
      * filter which encodes using the MIME Quoted Printable encoding.
      * See RFC 1521 for more information.
      */
    class output_filter_quoted_printable:
        public output_filter
    {
    public:
        /**
          * The destructor.
          */
        virtual ~output_filter_quoted_printable();

    private:
        /**
          * The constructor.
          * It is private on purpose, use the #create class method instead.
          *
          * @param deeper
          *     where to send the output of this filter.
          */
        output_filter_quoted_printable(const pointer &deeper);

    public:
        /**
          * The create class method is used to create new dynmically
          * allocated instances of this class.
          *
          * @param deeper
          *     where to send the output of this filter.
          */
        static pointer create(const pointer &deeper);

    protected:
        // See base class for documentation.
        ssize_t write(const void *data, size_t size);

    private:
        /**
          * The allow_international_characters instance variable is used
          * to remember whether or not internalional characters are
          * allowed (true) or to be quoted as unprintable (false).
          */
        bool allow_international_characters;

        struct glyph_t
        {
            unsigned char text;
            bool quote_it;
            int width;
            int cumulative;
        };

        enum { MAX_LINE_LEN = 76 }; // As per RFC 1521

        /**
          * The glyph instance variable is used
          * to remember
          */
        glyph_t glyph[MAX_LINE_LEN + 1];

        /**
          * The pos instance variable is used to remember the current
          * output position withing the glyph array.
          */
        int pos;

        /**
          * The eoln method is used to send the contents of the glyph
          * array to the deeper output stream.
          *
          * \param soft
          *     Either the array got too long (true),
          *     or a newline was seen (false).
          */
        void eoln(bool soft);

        /**
          * The newline_partial method is used to output a soft
          * end-of-line when the quoted line length exceeds the
          * MAX_LINE_LEN limit.
          */
        void eoln_partial();

        /**
          * The eoln_hard method is used to output a hard end-of-line,
          * when a newline character is seen.
          */
        void eoln_hard();

        /**
          * The default constructor.  Do not use.
          */
        output_filter_quoted_printable();

        /**
          * The copy constructor.  Do not use.
          *
          * @param rhs
          *     The right hand side of the initialization.
          */
        output_filter_quoted_printable(
            const output_filter_quoted_printable &rhs);

        /**
          * The assignment operator.  Do not use.
          *
          * @param rhs
          *     The right hand side of the assignment.
          */
        output_filter_quoted_printable &operator=(
            const output_filter_quoted_printable &rhs);
    };
}

#endif // LIBUNCIA_OUTPUT_FILTER_QUOTED_PRINTABLE_H
