//
// uncia - a big cat
// Copyright (C) 2008 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIBUNCIA_OUTPUT_FILTER_ROT13_H
#define LIBUNCIA_OUTPUT_FILTER_ROT13_H

#include <libuncia/output/filter.h>

namespace libuncia
{
    /**
      * The output_filter_rot13 class is used to represent a filter which
      * rotates letters 13 places.  It is its own inverse.
      *
      *
      * From Wikipedia (http://en.wikipedia.org/wiki/ROT13)...
      *
      * ROT13 ("rotate by 13 places", sometimes hyphenated ROT-13) is
      * a simple substitution cipher used in online forums as a means
      * of hiding spoilers, punchlines, puzzle solutions, and offensive
      * materials from the casual glance. ROT13 has been described as
      * the "Usenet equivalent of a magazine printing the answer to a
      * quiz upside down".[1] ROT13 is a variation of the Caesar cipher,
      * developed in ancient Rome.
      *
      * ROT13 is its own inverse; that is, to undo ROT13, the
      * same algorithm is applied, so the same action can be used
      * for encoding and decoding. The algorithm provides no real
      * cryptographic security and should never be used for such. It is
      * often cited as a canonical example of weak encryption. ROT13
      * has inspired a variety of letter and word games on-line, and is
      * frequently mentioned in newsgroup conversations.
      */
    class output_filter_rot13:
        public output_filter
    {
    public:
        /**
          * The destructor.
          */
        virtual ~output_filter_rot13();

    private:
        /**
          * The constructor.
          * It is private on purpose, use the #create class method instead.
          *
          * @param deeper
          *     where to send the filtered output.
          */
        output_filter_rot13(const pointer &deeper);

    public:
        /**
          * The create class method is used to create new dynmically
          * allocated instances of this class.
          *
          * @param deeper
          *     where to send the filtered output.
          */
        static pointer create(const pointer &deeper);

    protected:
        // See base class for documentation.
        ssize_t write(const void *data, size_t size);

    private:
        /**
          * The default constructor.  Do not use.
          */
        output_filter_rot13();

        /**
          * The copy constructor.  Do not use.
          *
          * @param rhs
          *     The right hand side of the initialization.
          */
        output_filter_rot13(const output_filter_rot13 &rhs);

        /**
          * The assignment operator.  Do not use.
          *
          * @param rhs
          *     The right hand side of the assignment.
          */
        output_filter_rot13 &operator=(const output_filter_rot13 &rhs);
    };
}

#endif // LIBUNCIA_OUTPUT_FILTER_ROT13_H
