//
// uncia - a big cat
// Copyright (C) 2008 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIBUNCIA_OUTPUT_FILTER_ROT47_H
#define LIBUNCIA_OUTPUT_FILTER_ROT47_H

#include <libuncia/output/filter.h>

namespace libuncia
{
    /**
      * The output_filter_rot47 class is used to represent
      * an output filter which performs the ROT47 encoding.
      *
      * See http://en.wikipedia.org/wiki/ROT13#ROT47
      *
      * "ROT47 is a derivative of ROT13 which, in addition to
      * scrambling the basic letters, also treats numbers and common
      * symbols. Instead of using the sequence A–Z as the alphabet,
      * ROT47 uses a larger set of characters from the common character
      * encoding known as ASCII. Specifically, all 7-bit printable
      * characters, excluding space, from decimal 33 '!' through 126
      * '~' are rotated by 47 positions, without special preserving
      * of case. The use of a larger alphabet is intended to produce
      * a more thorough obfuscation than that of ROT13, but because
      * ROT47 introduces numbers and symbols into the mix without
      * discrimination, it's much more obvious that the text has been
      * enciphered."
      */
    class output_filter_rot47:
        public output_filter
    {
    public:
        /**
          * The destructor.
          */
        virtual ~output_filter_rot47();

    private:
        /**
          * The constructor.
          * It is private on purpose, use the #create class method instead.
          *
          * @param deeper
          *     where to send the output of this filter.
          */
        output_filter_rot47(const pointer &deeper);

    public:
        /**
          * The create class method is used to create new dynmically
          * allocated instances of this class.
          *
          * @param deeper
          *     where to send the output of this filter.
          */
        static pointer create(const pointer &deeper);

    protected:
        // See base class for documentation.
        ssize_t write(const void *data, size_t size);

    private:
        /**
          * The default constructor.  Do not use.
          */
        output_filter_rot47();

        /**
          * The copy constructor.  Do not use.
          *
          * @param rhs
          *     The right hand side of the initialization.
          */
        output_filter_rot47(const output_filter_rot47 &rhs);

        /**
          * The assignment operator.  Do not use.
          *
          * @param rhs
          *     The right hand side of the assignment.
          */
        output_filter_rot47 &operator=(const output_filter_rot47 &rhs);
    };
}

#endif // LIBUNCIA_OUTPUT_FILTER_ROT47_H
