//
// uncia - a big cat
// Copyright (C) 2008 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIBUNCIA_OUTPUT_FILTER_UNEXPAND_H
#define LIBUNCIA_OUTPUT_FILTER_UNEXPAND_H

#include <libuncia/output/filter.h>

namespace libuncia
{
    /**
      * The output_filter_unexpand class is used to represent an output filter
      * which replaces sequences of spaces with tab charactares, where
      * this would result in fewer characters.
      */
    class output_filter_unexpand:
        public output_filter
    {
    public:
        /**
          * The destructor.
          */
        virtual ~output_filter_unexpand();

    private:
        /**
          * The constructor.
          * It is private on purpose, use the #create class method instead.
          *
          * @param deeper
          *     where to send the output of this filter.
          * @param leading_only
          *     whether to unexpand only leading white space (true), or
          *     all white space (false).
          * @param tab_width
          *     the width of each tab stop, in charactares.
          */
        output_filter_unexpand(const pointer &deeper, bool leading_only = true,
            int tab_width = 8);

    public:
        /**
          * The create class method is used to create new dynmically
          * allocated instances of this class.
          *
          * @param deeper
          *     where to send the output of this filter.
          * @param leading_only
          *     whether to unexpand only leading white space (true), or
          *     all white space (false).
          * @param tab_width
          *     the width of each tab stop, in characters.
          */
        static pointer create(const pointer &deeper, bool leading_only = true,
            int tab_width = 8);

    protected:
        // See base class for documentation.
        ssize_t write(const void *data, size_t size);

    private:
        /**
          * The deeper instance variable is used to remember whether to
          * unexpand only leading white space (true), or all white space
          * (false).
          */
        bool leading_only;

        /**
          * The deeper instance variable is used to remember the width
          * of each tab stop, in characters.
          */
        int tab_width;

        /**
          * The out_column instance variable is used to remember the
          * current output column.
          */
        int out_column;

        /**
          * The in_column instance variable is used to remember the
          * current input column.  This is can be greater than the
          * out_column in the presence of white space.
          */
        int in_column;

        /**
          * The default constructor.  Do not use.
          */
        output_filter_unexpand();

        /**
          * The copy constructor.  Do not use.
          *
          * @param rhs
          *     The right hand side of the initialization.
          */
        output_filter_unexpand(const output_filter_unexpand &rhs);

        /**
          * The assignment operator.  Do not use.
          *
          * @param rhs
          *     The right hand side of the assignment.
          */
        output_filter_unexpand &operator=(const output_filter_unexpand &rhs);
    };
}

#endif // LIBUNCIA_OUTPUT_FILTER_UNEXPAND_H
