//
// uncia - a big cat
// Copyright (C) 2008 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef UNCIA_ARGUMENT_SOURCE_H
#define UNCIA_ARGUMENT_SOURCE_H

#include <string>
#include <boost/shared_ptr.hpp>

/**
  * The argument_source class is used to represent an abstract source of
  * command line arguments.
  */
class argument_source
{
public:
    /**
      * The pointer typedef is provided to allow clients of this API
      * to be unaware of the actual details of the "smart" pointer
      * being used.  If it changes in future, all that is required
      * is a recompile.
      */
    typedef boost::shared_ptr<argument_source> pointer;

    /**
      * The destructor.
      */
    virtual ~argument_source();

protected:
    /**
      * The default constructor.
      */
    argument_source();

public:
    /**
      * The get method may be used to obtain the next command line
      * argument.
      *
      * @param token
      *     when to put the command line argument
      * @returns
      *     true if succeeds, false if no more available
      */
    virtual bool get(std::string &token) = 0;

    /**
      * The error method is used to report errors against specific
      * command line arguments (the one just returned by #next ).
      *
      * @param text
      *     The text of the error message.
      *
      * @note
      *     (No printf-like formatting is available.)
      */
    virtual void error(const char *text) const = 0;

    /**
      * The compare class method is used to compare a command line string
      * with a formal spec of the option, to see if they compare equal.
      *
      * The actual is case-insensitive.  Uppercase in the formal means a
      * mandatory character, while lower case means optional.  Any number of
      * consecutive optional characters may be supplied by actual, but none
      * may be skipped, unless all are skipped to the next non-lower-case
      * letter.
      *
      * The underscore (_) is like a lower-case minus, it matches "", "-"
      * and "_".
      *
      * The "*" in a pattern matches everything to the end of the line,
      * anything after the "*" is ignored.  The rest of the line is pointed
      * to by the "partial" variable as a side-effect (else it will be 0).
      * This rather ugly feature is to support "-I./dir" type options.
      *
      * A backslash in a pattern nominates an exact match required, case
      * must matche excatly here.  This rather ugly feature is to support
      * "-I./dir" type options.
      *
      * For example: "-project" and "-P' both match "-Project", as does
      * "-proJ", but "-prj" does not.
      *
      * For example: "-devDir" and "-d_d' both match "-Development_Directory",
      * but "-dvlpmnt_drctry" does not.
      *
      * For example: to match include path specifications, use a pattern
      * such as "-\\I*", and the partial global variable will have the path
      * in it on return.
      */
    static bool compare(const char *formal, const char *actual);

private:
    /**
      * The copy constructor.  Do not use.
      *
      * @param rhs
      *     The right hand side of the initialization.
      */
    argument_source(const argument_source &rhs);

    /**
      * The assignment operator.  Do not use.
      *
      * @param rhs
      *     The right hand side of the assignment.
      */
    argument_source &operator=(const argument_source &rhs);
};

#endif // UNCIA_ARGUMENT_SOURCE_H
