//
//      uncia - a big cat
//      Copyright (C) 2008 Peter Miller
//
//      This program is free software; you can redistribute it and/or
//      modify it under the terms of the GNU General Public License,
//      version 3, as published by the Free Software Foundation.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program. If not, see
//      <http://www.gnu.org/licenses/>.
//

#include <iostream>
#include <string>

#include <libuncia/input.h>
#include <libuncia/output.h>
#include <libuncia/progname.h>
#include <libuncia/sizeof.h>

#include <uncia/argument_source/gnu.h>
#include <uncia/argument_source/include.h>
#include <uncia/argument_source/main.h>
#include <uncia/gram.yacc.h>
#include <uncia/lex.h>


struct table_t
{
    const char *name;
    int value;
};

static const table_t table[] =
{
    { "-Base_64", OPTION_BASE64 },
    { "-Carriage_Return_Line_Feed", OPTION_CRLF },
    { "-COLumn_ReMove", OPTION_COLRM },
    { "-Double_Space", OPTION_DOUBLE_SPACE },
    { "-Down_Case", OPTION_DOWNCASE },
    { "-Echo", OPTION_ECHO },
    { "-Escaped_NewLines", OPTION_ESCAPE_NEWLINE },
    { "-EXpand", OPTION_EXPAND },
    { "-FOld", OPTION_FOLD },
    { "-Fold_Spaces", OPTION_FOLD_SPACES },
    { "-Gnu_UnZip", OPTION_GUNZIP },
    { "-Gnu_Zip", OPTION_GZIP },
    { "-Hash_Comments", OPTION_HASH_COMMENTS },
    { "-HEad", OPTION_HEAD },
    { "-Hexadecimal_Dump", OPTION_HEXDUMP },
    { "-Lower_Case", OPTION_DOWNCASE },
    { "-number_nonBlank", OPTION_NUMBER_NONBLANK },
    { "-Number", OPTION_NUMBER },
    { "-Output", OPTION_OUTPUT },
    { "-PAste_Delimited", OPTION_PASTE_DELIM },
    { "-PAste_Delimiter", OPTION_PASTE_DELIM },
    { "-PAste", OPTION_PASTE },
    { "-PREfix", OPTION_PREFIX_ADD },
    { "-Quoted_Printable", OPTION_QUOTED_PRINTABLE },
    { "-Remove_Prefix", OPTION_PREFIX_REMOVE },
    { "-Remove_SUffix", OPTION_SUFFIX_REMOVE },
    { "-REVerse-Lines", OPTION_REVERSE_LINES },
    { "-REVerse", OPTION_REVERSE_CHARS },
    { "-Rotate13", OPTION_ROT13 },
    { "-Rotate47", OPTION_ROT47 },
    { "-Show_All", OPTION_SHOW_ALL },
    { "-Show_Ends", OPTION_SHOW_ENDS },
    { "-Show_Non_Printing", OPTION_SHOW_NONPRINTING },
    { "-Show_Tabs", OPTION_SHOW_TABS },
    { "-SORT", OPTION_SORT },
    { "-Squeeze_Blank", OPTION_SQUEEZE_BLANK },
    { "-SUffix", OPTION_SUFFIX_ADD },
    { "-TAil", OPTION_TAIL },
    { "-TEE", TEE },
    { "-UnExpand_All", OPTION_UNEXPAND_ALL },
    { "-UnExpand", OPTION_UNEXPAND },
    { "-UNique_Count", OPTION_UNIQUE_COUNT },
    { "-UNique", OPTION_UNIQUE },
    { "-Unix_to_Unix_Decode", OPTION_UUDECODE },
    { "-Unix_to_Unix_Encode", OPTION_UUENCODE },
    { "-Upper_Case", OPTION_UPCASE },
    { "-VERSion", VERSION },
    { "-Yydebug", OPTION_gram_DEBUG },
};


static argument_source::pointer src;
static int number_of_errors;


void
lex_open(int argc, char **argv)
{
    src =
        argument_source_gnu::create
        (
            argument_source_include::create
            (
                argument_source_main::create(argc, argv)
            )
        );
    number_of_errors = 0;
}


static bool
is_a_number(const std::string &s, long &result)
{
    const char *cp = s.c_str();
    char *ep = 0;
    long n = strtol(cp, &ep, 0);
    if (ep == cp || *ep)
        return false;
    result = n;
    return true;
}


int
gram_lex()
{
    if (!src)
        return 0;
    std::string s;
    if (!src->get(s))
        return 0;
    if (is_a_number(s, gram_lval.lv_integer))
        return INTEGER_CONSTANT;
    if (s[0] == '-')
    {
        if (s[1] == '\0')
            return STDIO;

        typedef std::list<const table_t *> hits_t;
        hits_t hits;
        for (const table_t *tp = table; tp < ENDOF(table); ++tp)
        {
            if (argument_source::compare(tp->name, s.c_str()))
                hits.push_back(tp);
        }
        switch (hits.size())
        {
        case 0:
            return OPTION_UNKNOWN;

        case 1:
            return hits.back()->value;

        default:
            //
            // the option isn't ambiguous if they all have the
            // same value (it could be present a two different
            // spellings).
            //
            {
                bool same = true;
                for
                (
                    hits_t::const_iterator it = hits.begin();
                    it != hits.end();
                    ++it
                )
                {
                    if ((*it)->value != hits.front()->value)
                    {
                        same = false;
                        break;
                    }
                }
                if (same)
                    return hits.back()->value;
            }

            //
            // Complain about ambiguous options.
            //
            {
                std::string message;
                for
                (
                    hits_t::const_iterator it = hits.begin();
                    it != hits.end();
                    ++it
                )
                {
                    if (!s.empty())
                        message += ", ";
                    message += (*it)->name;
                }
                message = "option " + s + " is ambiguous (" + message + ")";
                gram_error(message.c_str());
                exit(1);
            }
        }
    }
    if (s == "(" || s == "{" || s == "[")
        return LP;
    if (s == ")" || s == "}" || s == "]")
        return RP;
    gram_lval.ssp = new std::string(s);
    return STRING_CONSTANT;
}


void
gram_error(const char *text)
{
    ++number_of_errors;
    if (!src)
    {
        std::cerr << libuncia::progname_get() << ": " << text << std::endl;
        exit(1);
    }
    src->error(text);
}


void
lex_close()
{
    if (number_of_errors > 0)
    {
        std::cerr << libuncia::progname_get() << ": found " << number_of_errors
            << " fatal error" << (number_of_errors == 1 ? "" : "s")
            << std::endl;
        exit(1);
    }
    src.reset();
}
